#!/usr/bin/env python
import ast
import dis
import sys
import types
from astoptimizer import Config, parse_ast, optimize_ast, compile_ast
from optparse import OptionParser

def dump_bytecode(code):
    dis.dis(code)
    for const in code.co_consts:
        if not isinstance(const, types.CodeType):
            continue
        print("")
        print("%s:" % const)
        dump_bytecode(const)

def main():
    parser = OptionParser(usage="%prog [options] code [line2 line3 ...]")
    parser.add_option("-f", "--file", metavar="SCRIPT",
        help="execute specified script",
        type="str", default=None)
    parser.add_option("-k",
        help="Don't remove dead code",
        action="store_true", default=False)
    parser.add_option("-x",
        help="Enable experimental variables feature",
        action="store_true", default=False)
    parser.add_option("-n",
        help='Disable all configuration features',
        action="store_true", default=False)
    parser.add_option("-e", "--enable", metavar="FEATURES",
        help='Enable configuration features FEATURES (default: builtin_funcs and pythonbin)',
        action="store", type="str", default="builtin_funcs,pythonbin")
    parser.add_option("-a", "--all",
        help="Enable all optimizations",
        action="store_true", default=False)
    parser.add_option("--str-limit",
        help="Maximum length of strings in bytes or characters",
        action="store", type="int", default=None)
    parser.add_option("--tuple-limit",
        help="Maximum length of frozenset, set and tuple",
        action="store", type="int", default=None)
    parser.add_option("--no-exec",
        help="Maximum length of frozenset, set and tuple",
        action="store_true", default=False)
    options, args = parser.parse_args()
    if not options.file and not args:
        parser.print_help()
        exit(1)

    if options.file:
        filename = options.file
        with open(filename, 'r') as fp:
            code_str = fp.read()
    else:
        filename = None
        code_str = '\n'.join(args)

    config = Config()
    if options.all:
        config.enable_all_optimizations()
    if not options.n:
        for feature in options.enable.split(','):
            config.enable(feature) #'builtin_funcs', 'pythonenv')
    if options.x:
        config.use_experimental_vars = True
    if options.k:
        config.remove_dead_code = False
    if options.str_limit is not None:
        config.max_string_length = options.str_limit
    if options.tuple_limit is not None:
        config.max_tuple_length = options.tuple_limit
    print("Features: %s" % ', '.join(sorted(config.features)))
    print("Experimental varaibles? %s" % config.use_experimental_vars)
    print("Remove code? %s" % config.remove_dead_code)
    print("Limits: string=%s, tuple=%s" % (config.max_string_length, config.max_tuple_length))
    print("")

    tree = parse_ast(code_str)
    print("Original AST:")
    print(ast.dump(tree))
    print("")

    print("Original bytecode:")
    try:
        code = compile(code_str, "<string>", "exec")
    except SyntaxError:
        print("*** SYNTAX ERROR ***")
    else:
        dump_bytecode(code)
    print("")

    tree = optimize_ast(tree, config, filename)
    print("Optimised AST:")
    print(ast.dump(tree))
    print("")

    print("Optimized bytecode:")
    try:
        code = compile_ast(tree)
    except SyntaxError:
        print("*** SYNTAX ERROR ***")
    else:
        dump_bytecode(code)
        print("")

        if not options.no_exec:
            print("Execute code:")
            exec(code)

if __name__ == "__main__":
    main()

